////////////////////////////////////////////////////////
// Copyright (c) M.Iakobovski 11.06.2007 Version v.02 //
// Generation pseudorandom numbers sequence           //
// with direct access                                 //
////////////////////////////////////////////////////////

#include "stdio.h"
#include "stdlib.h"
#include "memory.h"
#include "bint.h"

int bint_init_BINT_UL(LPBINT pi, int nbits, UL c)
{
	nbits=BINT_BITS; // !! ##   10.06.2007  255 

// printf("bint_init_BINT_UL: nbits=%d\n",nbits);

 pi->nbits=nbits;
 pi->nparts=2*((pi->nbits+USBITS-1)/USBITS);

 pi->a=malloc(sizeof(UL)*pi->nparts);
 if(!pi->a)
	{
	 printf("bint_init_BINT_UL: Error alloc pi->a mas (%d parts)\n",pi->nparts);
	 exit(1);
	}

 bint_set_BINT_UL(pi, c);

 return 0;
}


int bint_set_BINT_UL(LPBINT pi, UL c)
{
 UL c0,c1;

 memset(pi->a,0,sizeof(UL)*pi->nparts);

 c0=c&USMAX;
 c1=c>>USBITS;

 pi->a[0]=c0;
 pi->a[1]=c1;

 return 0;
}

int bint_isbit(LPBINT pi, int i)
{
 return (pi->a[i/USBITS]&(1<<(i%USBITS)))!=0;
}

int bint_add_BINT_UL(LPBINT pi, UL c)
{
 int i;
 UL a,p=0;
 UL c0,c1;

 c0=c&USMAX;
 c1=c>>USBITS;

 for(p=c0,i=0;p && (i<pi->nparts);i++)
	{
	 a=pi->a[i]+p;

	 if(a>USMAX)	p=1;
	 else			p=0;

	 a&=USMAX;

	 pi->a[i]=a;
	}

 for(p=c1,i=1;p && (i<pi->nparts);i++)
	{
	 a=pi->a[i]+p;

	 if(a>USMAX)	p=1;
	 else			p=0;

	 a&=USMAX;

	 pi->a[i]=a;
	}

 return p;
}

int bint_mult_BINT_UL(LPBINT pi, UL c)
{
 int i,pn=pi->nparts;
 UL a,b,p;
 UL c0,c1;
 UL b0,b1;
 static UL *r=NULL;
 unsigned long *pa,*pb;

// printf("pi="); bint_print(stdout,pi); 
// printf("\n!!! pi->a[0]=%u\n",pi->a[0]);

 if(!r)
	{
//		printf("pi->nparts=%d\n",pi->nparts);
	r=(UL *)malloc(sizeof(UL)*pi->nparts);
	if(!r)
		{
		 printf("bint_mult_BINT_UL: Error alloc r mas (%d parts)\n",pi->nparts);
		 exit(1);
		}
	}

 if(c==(USMAX+1)) // 0x10000 -      
 {
	pa=pi->a+pn-1;
	pb=pa-1;
	p=*pa;

	for(i=pn-1;i;i--)
		*(pa--)=*(pb--);
	*pa=0;
/*
	pa=pi->a;
	p=pa[pn-1];

	for(i=pn-1;i;i--)
		pa[i]=pa[i-1];
	pa[0]=0;
*/	return p;
 }

// printf("!!! pi->a[0]=%u\n",pi->a[0]);

//		printf("pi->nparts=%d\n",pi->nparts);
 memset(r,0,sizeof(UL)*pi->nparts);
// printf("!!! pi->a[0]=%u\n",pi->a[0]);

 c0=c&USMAX;
 c1=c>>USBITS;

 p=0;

#if 1 //   
 pa=pi->a;
 pb=r;

// printf("!!! pi->a[0]=%u\n",pi->a[0]);

// printf("\n\n\n *pa=%u %u\n",*pa,pi->a[0]);

 if(c0)
 for(p=0,i=pi->nparts-1;i;i--)
	{
	 b=(*pa++)*c0;
//	 printf("b=%u\n",b);

	 b0=b&USMAX;
	 b1=b>>USBITS;

	 a=b0+p;

	 *pb++=a&USMAX;
	 p=b1+(a>>USBITS);
//	 printf("%3d %u\n",i,p);
	}
#else
 if(c0)
 for(p=0,i=0;i<pi->nparts;i++)
	{
	 b=pi->a[i]*c0;

	 b0=b&USMAX;
	 b1=b>>USBITS;

	 a=b0+p;

	 r[i]=a&USMAX;
	 p=b1+(a>>USBITS);
	}
#endif

 if(c1)
 for(p=0,i=0;i<pi->nparts;i++)
	{
	 b=pi->a[i]*c1+r[i+1]+p;

	 r[i+1]=b&USMAX;
	 p=b>>USBITS;
	}

 memcpy(pi->a,r,sizeof(UL)*pi->nparts);
 return p;
}

