////////////////////////////////////////////////////////
// Copyright (c) M.Iakobovski 11.06.2007 Version v.02 //
// Generation pseudorandom numbers sequence           //
// with direct access                                 //
////////////////////////////////////////////////////////

#include "stdio.h"
#include "stdlib.h"
#include "memory.h"
#include "prand.h"

#define if0 if(0)
#define qq if(0)

int init_vect(VECT v, int nbits);
void		sort_mas(int *a, int n);

 unsigned long count1,count2;

int init_vect(VECT v, int nbits)
 {
 v->offset=0;
 v->n=nbits;
 
 v->r=malloc(2*v->n+1);

 if(!v->r)
	{
	 printf("init_vect: Error alloc v->r (%d bits)\n",v->n);
	 exit(0);
	}

 memset(v->r,0,2*v->n+1);

 return 0;
 }

 void prand_init(LPPRAND pr, int k, int bits_of_result) //    
{
 int nbits;

 pr->bits_of_result=bits_of_result;

 pr->bcompressed=0;

 switch(k)
 {
 case  255: 

	 nbits=255; 
 pr->nbase=5;
 pr->base[0]=0;
 pr->base[1]=3; // 3
 pr->base[2]=7; // 7
 pr->base[3]=31; // 31
 pr->base[4]=255; 
 pr->bcompressed=1;
 break;

 default:
	 printf("prand_init: Unsupported k=%d\n",k);
	 exit(1);
 }
 //printf("prand_init :bcompressed=%d\n",pr->bcompressed);

 init_vect(&(pr->r),nbits);

 pr->r.r[0]=1; // It seems f(x)=x :).
}

unsigned long prand(LPPRAND pr) //      
{
 int b=0,n=pr->r.n,k,o;
 unsigned long res;
 char *r=pr->r.r;
 int of=pr->r.offset;
 int nb;
 int *pb;

 for(res=0,k=pr->bits_of_result;k;k--)
	{
	 if(of)	of--;
	 else	of=n-1;

	 res<<=1;

	 if(r[of]) //     x^(n-1)
		{
		res++;
		for(nb=pr->nbase-2,pb=pr->base+1;nb;nb--)
			{
			o=of+(*(pb++));//pr->base[i];
			if(o>=n)
				o-=n;

			r[o]^=1;
			}
		}
	}

 pr->r.offset=of;

// {	 static int ncall; printf("prand%5d = %7lu\n",ncall++,res); }
//	printf("%u\n",pr->r.r[0]);

 return res;
}

unsigned long prand_set_c(LPPRAND pr, LPBINT pbi0); // c     bi
void qvadr_vect_c(CVECT *cv0, CVECT *cp, CVECT *cv1); // cv1 = (cv0*cv0) mod cp
void att_to_vect_c(CVECT *cv, CVECT *cp, int i0); // cv1 += x^i0 mod cp
void add_to_mas(int *mas, int *nmas, int i0);
void mult_vect_c( VECT pr0, CVECT *cx0, CVECT *cp, VECT pr1); // r1 = (r0*cx0) mod cp
void mult_vect_c_precomp( VECT pr0, CVECT *x0, int ix0/*CVECT *cx0*/, CVECT *cp, VECT pr1); // r1 = (pr0*cx0) mod cp

void mult_vect_c_precomp( VECT pr0, CVECT *x0, int ix0/*CVECT *cx0*/, CVECT *cp, VECT pr1) // r1 = (pr0*cx0) mod cp
{
 int i,j,l,jb,n=cp->m,nn=pr1->n,*p,ln;
 CVECT *cx0=x0+ix0;
 char *r=pr1->r;

#define pt(lll) if(0)printf("pt ^r1[%d] j=%d ix0=%d\n",lll,j,ix0)
#define pt0(lll) if(0)printf("pt ^r1[%d] j=%d i=%d ix0=%d\n",lll,j,i,ix0)

 pr1->offset=0;

 memset(r,0,n); 

	 for(j=n;j;j--)
		if(pr0->r[j-1])
			 for(i=0;i<cx0->n;i++)
				{
				l=j-1+cx0->a[i]; //   x^(l)

				 if(l>=nn)
				 {
					p=cp->a;
					ln=l-n;
					for(jb=cp->n-1;jb;jb--)
					{
						r[ln+(*p++)]^=1;
					}
				 }
				 else
					{
						r[l]^=1;
					}
				}
}

void add_to_mas(int *a, int *pn, int b)
{
 int i;
#define n (*pn)

 for(i=0;i<n;i++)
	 if(a[i]==b)
		{
		 a[i]=a[n-1];
		 n--;
		 return;
		}

 a[i]=b;
 n++;

#undef n
}

void add_to_vect_c(CVECT *cv, CVECT *cp, int i0) // cv1 += x^i0 mod cp
{
 int i;

 if(i0<cp->m)
	 add_to_mas(cv->a,&(cv->n),i0);
 else
	 for(i=0;i<cp->n;i++)
		 if(cp->a[i]<cp->m)
			 add_to_mas(cv->a,&(cv->n),cp->a[i]+i0-cp->m);
}

void qvadr_vect_c(CVECT *cv0, CVECT *cp, CVECT *cv1) // cv1 = (cv0*cv0) mod cp
{
 int i;

 for(i=0;i<cv0->n;i++)
	 add_to_vect_c(cv1,cp,2*cv0->a[i]);
}

int	init_vect_c(CVECT *cv, int m)
{
	cv->n=0;
	cv->m=m;
	return 0;
}


void out_pol(int *a, int  n)
{
 int i,k=0;

 for(i=n-1;i>=0;i--)
		{
		 if(k)
			 printf("+");

		 k=1;
		 if(a[i]==0)	 printf("1");
		 else
		 if(a[i]==1)	 printf("x");
		 else
         			 printf("x^%d",a[i]);
		}
}


unsigned long prand_set_c(LPPRAND pr, LPBINT pbi0) // c     bi
{
 static RPOLE r0, r1; // r -  
 int n=pr->r.n,i;
 static BINT b;
 static CVECT *cx0,cp; // cx0 -   x ; 
 RPOLE *pr0,*pr1,*ptmp;

 //     ()
 //      i*bits_of_result
 //   i    (i*bits_of_result)  ((i+1)*bits_of_result-1)

 if(!cx0)
	{
 //   printf("\np(x)=");	out_pol(pr->base,pr->nbase);	printf("\n");

	cp.n=pr->nbase;
	cp.m=n;
	memcpy(cp.a,pr->base,sizeof(int)*cp.n);

	cx0=(CVECT*)malloc(sizeof(CVECT)*(n+1));
	if(!cx0)
		{
			printf("prand_set_c: Error alloc cx0 (%d)\n",n);
			exit(1);
		}

	memset(cx0,0,sizeof(CVECT)*(n+1));

	bint_init_BINT_UL(&b,pbi0->nbits+pr->bits_of_result+1,0);

	init_vect(&r0,n);
	init_vect(&r1,n);

	init_vect_c(cx0,n);
	cx0[0].a[0]=1; // It seems f(x)=x :). //   x : x^1
	cx0[0].n=1;

#define if1 if(1)

	// out_pol(pr->base,pr->nbase);

	i=0;
	qq if1 {printf("\nx0[2^%3d] ",i); out_pol(cx0[i].a,cx0[i].n); printf("");}
	for(i=0;i<n;i++) //      
		{
		init_vect_c(cx0+i+1,n);
		qvadr_vect_c(cx0+i, &cp, cx0+i+1); // x1 = (x0*x0) mod p

//		sort_mas((cx0+i+1)->a,(cx0+i+1)->n);

		qq if1 {printf("\nx0[2^%3d] ",i+1); out_pol(cx0[i+1].a,cx0[i+1].n);  printf("");}
		}
	qq printf("\n");
	}

// for(i=0;i<pbi0->nparts;i++)	 b.a[i]=pbi0->a[i];
 memcpy(b.a,pbi0->a,sizeof(unsigned long)*pbi0->nparts);

//printf("b="); bint_print(stdout,&b); printf("\n");

bint_mult_BINT_UL(&b,pr->bits_of_result);

//printf("b="); bint_print(stdout,&b); printf("\n");

 memset(r0.r,0,n);
 memset(r1.r,0,n);

 r0.offset=0;	
 r0.r[0]=1; // It seems f(x)=1 :). //    1 : x^0

 pr0=&r0;
 pr1=&r1;


 for(i=0;i<b.nbits;i++) //      
	{
	if(bint_isbit(&b,i))
		{
		mult_vect_c_precomp( pr0, cx0,i%n, &cp, pr1); // r1 = (r0*x0) mod p lrnd32direct nbits=255 k= 100000 = 11.485 sec r=1888093936   
		ptmp=pr0;
		pr0=pr1;
		pr1=ptmp;
		}

    //if0 if(bi.a[0]==16) {printf("\nx0 =x1  = \n"); put_vect(&x0); 	out_vecti(&x0); printf("\n");}
	}

 pr->r.offset=0;
// memcpy(pr->r.r,r0.r,n);
 memcpy(pr->r.r,pr0->r,n);

 return prand(pr);
}

void qvadr_vect(VECT a, VECT p, VECT r) // r = (a*a) mod p
{
 static RPOLE b,c;
 int i,m=a->n;

 if(!b.n)
	{
	init_vect(&b,m);
	init_vect(&c,m);
	}

 memset(b.r,0,2*m+1);
 memset(c.r,0,2*m+1);

 for(i=0;i<=m;i++)
	 b.r[2*i]=a->r[i];

 if0 {printf("b= "); put_vect(&b); 	out_vecti(&b); printf("\n");}
 if0 {printf("p= "); put_vect(p); 	out_vecti(p); printf("\n");}
 devide_vect( &b, p, &c, r); // b=pc+r 
 if0 {printf("c= "); put_vect(&c); 	out_vecti(&c); printf("\n");}
 if0 {printf("r= "); put_vect(r); 	out_vecti(r); printf("\n");}
}

void mult_vect(VECT a1, VECT a2, VECT p, VECT r) // r = (a1*a2) mod p
{
 static RPOLE b,c;
 int i,j,m=a1->n;

 if(!b.n)
	{
	init_vect(&b,m);
	init_vect(&c,m);
	}

 memset(b.r,0,2*m+1);
 memset(c.r,0,2*m+1);

 for(i=0;i<=m;i++)
	 if(a1->r[i])
		for(j=0;j<=m;j++)
			b.r[i+j]=(b.r[i+j]+a2->r[j])%2;

 devide_vect( &b, p, &c, r); // a=bc+r 
}

void out_vecti(VECT a)
{
 int i,k=0;

 for(i=2*a->n;i>=0;i--)
	 if(a->r[i])
		{
		 if(k)
			 printf("+");
		 k=1;
		 if(i==0)	 printf("1");
		 else
		 if(i==1)	 printf("x");
		 else
		         	 printf("x^%d",i);
		}
}

unsigned int devide_vect(VECT a, VECT b, VECT c, VECT r) // a=bc+r 
{
 int i,j,ret,sta,m=a->n;

 static int stb;
 static int *dvd;
 if(!dvd)
	{
	 dvd=(int *)malloc((2*m+1)*sizeof(int));
	 if(!dvd)
	 {
		 printf("devide_vect: error alloc dvd\n");
		 exit(0);
	 }
	 for(stb=0,i=0;i<=2*m;i++)if(b->r[i])dvd[stb++]=i;
	}

 ret=0;
 for(sta=0,i=0;i<=2*m;i++)if(a->r[i])sta=i; // -    

 memset(c->r,0,2*m);
 memcpy(r->r,a->r,2*m);

 if(dvd[stb-1]>sta)
	 ret=1;
 else
  {
   for(i=sta;i>=dvd[stb-1];i--)
	{
	  if(r->r[i])
		{
		for(j=0;j<stb;j++)
			r->r[i-(dvd[stb-1]-dvd[j])] = 1-r->r[i-(dvd[stb-1]-dvd[j])]; // ta = ta xor tb
		c->r[i-dvd[stb-1]]=1;
		}
	}
	for(i=0;i<m;i++)if(r->r[i])break;

	if(i<m)
		ret=1;
  }

  if(0)
//  if(!ret)
	{
	printf("\n devide_vect : ");
	out_vecti(a);  printf(" = ( "); 
	out_vecti(b);  printf(" ) *  ( "); 
	out_vecti(c);  printf(" ) + "); 
	out_vecti(r);  printf(" ;\n"); 
	}
 return ret;
}




void put_vect(VECT b)
{
  int i,m=b->n;//,j;
#if 1
  printf("off=%d\t",b->offset);
  for(i=m-1;i>=0;i--)
	printf("%d",b->r[(i+1*b->offset)%m]);
  printf(" : ");
#else
  j=0;
  for(i=0;i<m;i++)
	{
	  j*=2;
	  j+=b->r[m-i-1];
	}
  printf("%08x ",j);

#endif
}

